import json
import os
import tempfile
import zipfile
from typing import Union

import pandas as pd
import requests

from io import StringIO

from glm_met import glm_met

from . import settings


class MetData(glm_met.MetData):
    """
    Class for meteorological data and its associated metadata.

    Used  to store meteoroligcal data downloaded from the
    SILO API.

    Attributes
    ----------
    metadata : (dict)
        Metadata associated with the meteorological data.
    data : (pd.DataFrame)
        Meteorological data in a Pandas DataFrame.
    """

    def __init__(self, metadata: dict, data: pd.DataFrame):
        self.metadata = metadata
        self.data = data


class Silo(glm_met.GlmMet):
    """
    Class for retrieving and processing historical meteorological data from
    the SILO API.

    Examples
    --------
    >>> import os
    >>> import glm_met.silo.silo as silo
    >>> s = silo.Silo(
    ...            location=(116.6, -32.17),
    ...            date_range=("20220101", "20220131"),
    ...            met_data=None,
    ...            format="csv",
    ...            comment="rxel",
    ...            username="test@email.com",
    ...            api="data_drill"
    ...        )
    >>> s.get_variables(request_settings=None)
    >>> s.write_met(path="silo-data")

    After successful call to `get_variables()`, the `met_data` attribute
    of the `Silo` type object, `s`, is a `MetData` type object.
    This object has two attributes: a metadata dict and a DataFrame of
    met data values.

    >>> metadata = s.met_data.metadata
    >>> met_data_df = s.met_data.data
    """

    def __init__(
        self,
        location: Union[int, tuple[float, float]],
        date_range: tuple[str, str],
        met_data: Union[None, MetData],
        comment: Union[None, str],
        username: str,
        api: str,
    ):
        """
        Initialise object to download data from the SILO API.

        Parameters
        ----------
        location : Union[int, tuple[float, float]]
            Station number for the patch point API or
            latitude and longitude tuple for the drill down grid
            API.
        date_range : tuple[str, str]
            Start and end dates for the data retrieval
            (in "YYYYMMDD" format).
        met_data : Union[None, MetData]
            Attribute to store meteorological data downloaded from SILO API.
        comment :  str
            String codes representing meteorological variables to download.
            See list of variables here: `https://www.longpaddock.qld.gov.au/silo/about/climate-variables/`
            For example, `"rxel"` will download rainfall, max temperature,
            evaporation, and Morton's shallow lake evaporation.
        username : str
            Email address that must be provided.
        api :  str
            Either "drill_down" for the retrieving point-like data from the gridded
            SILO product or "patch_point" for station data.

        """
        self.location = location
        self.date_range = date_range
        self.met_data = met_data
        self.format = "csv"
        self.comment = comment
        self.username = username
        self.api = api

    def get_variables(self, request_settings: Union[None, dict]) -> MetData:
        """
        Get variables from the SILO API and store them
        in the `met_data` attribute.

        Parameters
        ----------
        request_settings : Union[None, dict]
            Dictionary object of extra settings to pass in as query parameters to
            the SILO API.
        """

        if self.api == "data_drill":
            # location should be tuple - lat and lon
            if type(self.location) == tuple:
                if self.format == "json" or self.format == "csv":
                    payload = {
                        "lon": self.location[0],
                        "lat": self.location[1],
                        "start": self.date_range[0],
                        "finish": self.date_range[1],
                        "format": self.format,
                        "comment": self.comment,
                        "username": self.username,
                        "password": "apirequest",
                    }

                else:
                    payload = {
                        "lon": self.location[0],
                        "lat": self.location[1],
                        "start": self.date_range[0],
                        "finish": self.date_range[1],
                        "format": self.format,
                        "username": self.username,
                        "password": "apirequest",
                    }

                # if user has supplied extra query parameters, append them here.
                if request_settings:
                    payload = {**payload, **request_settings}

                r = requests.get(settings.data_drill_api, params=payload)

                df = pd.read_csv(StringIO(r.text))

                # make meta data dict
                metadata_col = df.loc[:, "metadata"]
                elevation = metadata_col.iloc[0].split("= ")[1].split(" m")[0]
                metadata = {
                    "longitude": self.location[0],
                    "latitude": self.location[1],
                    "start": self.date_range[0],
                    "finish": self.date_range[1],
                    "api": self.api,
                    "format": self.format,
                    "comment": self.comment,
                    "elevation": elevation,
                }

                df = df.drop(columns=["metadata"])
                self.met_data = MetData(metadata=metadata, data=df)

        elif self.api == "patch_point":
            # location should be int type - station number
            if type(self.location) == int:
                if self.format == "json" or self.format == "csv":
                    payload = {
                        "station": self.location,
                        "start": self.date_range[0],
                        "finish": self.date_range[1],
                        "format": self.format,
                        "comment": self.comment,
                        "username": self.username,
                        "password": "apirequest",
                    }

                else:
                    payload = {
                        "station": self.location,
                        "start": self.date_range[0],
                        "finish": self.date_range[1],
                        "format": self.format,
                        "username": self.username,
                        "password": "apirequest",
                    }

                # if user has supplied extra query parameters, append them here.
                if request_settings:
                    payload = {**payload, **request_settings}

                r = requests.get(settings.patch_point_api, params=payload)

                df = pd.read_csv(StringIO(r.text))

                # make meta data dict
                metadata_col = df.loc[:, "metadata"]
                elevation = metadata_col.iloc[3].split("= ")[1].split(" m")[0]
                lat = metadata_col.iloc[1].split("= ")[1]
                lon = metadata_col.iloc[2].split("= ")[1]
                name = metadata_col.iloc[0].split("=")[1]

                metadata = {
                    "station": self.location,
                    "station_name": name,
                    "longitude": lon,
                    "latitude": lat,
                    "start": self.date_range[0],
                    "finish": self.date_range[1],
                    "api": self.api,
                    "format": self.format,
                    "comment": self.comment,
                    "elevation": elevation,
                }

                df = df.drop(columns=["metadata"])
                self.met_data = MetData(metadata=metadata, data=df)

    def write_met(self, path: str) -> None:
        """
        Save meteorological data and its metadata to a zip file.

        Parameters
        ----------
        path : str
            Path to the directory where the zip file should be saved.
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            self.met_data.data.to_csv(
                os.path.join(tmpdir, "met_raw.csv"), index=False
            )

            with open(
                os.path.join(tmpdir, "met_raw_metadata.json"), "w"
            ) as dst:
                json.dump(self.met_data.metadata, dst)

            data_fpath = os.path.join(tmpdir, "met_raw.csv")
            metadata_fpath = os.path.join(tmpdir, "met_raw_metadata.json")

            files = [data_fpath, metadata_fpath]

            with zipfile.ZipFile(
                os.path.join(path, "met_raw.zip"), mode="w"
            ) as z_dst:
                for file in files:
                    z_dst.write(file, arcname=file.split("/")[-1])

    def convert_to_glm(self) -> None:
        """Convert weather data to GLM format."""
        pass

    def write_glm_met(self) -> None:
        """Save weather data in GLM format."""
        pass



nasa_power_api_url = "https://power.larc.nasa.gov/api/temporal/hourly/point"

# get parameter names from: https://power.larc.nasa.gov/#documentation
hourly_glm_default = [
    "ALLSKY_SFC_SW_DWN",  # W/m^2
    "CLOUD_AMT",  # %
    "T2M",  # degC
    "RH2M",  # %
    "WS2M",  # m/s
    "PRECTOTCORR",  # mm/hour
]



import pytest
import tempfile
import os
import pandas as pd
import zipfile

import glm_met.silo.silo as silo

def test_silo_init():
    """Test initialise Silo object"""
    tmp_silo = silo.Silo(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None, 
        comment="rxel",
        username="testing-glm-met@pytest.com", 
        api="data_drill"
    )

    assert isinstance(tmp_silo, silo.Silo)


def test_silo_drill_down():
    """Test download of data from SILO drill down"""
    tmp_silo = silo.Silo(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None, 
        comment="rxel",
        username="testing-glm-met@pytest.com", 
        api="data_drill"
    )

    tmp_silo.get_variables(request_settings=None)

    assert isinstance(tmp_silo.met_data.metadata, dict) and isinstance(tmp_silo.met_data.data, pd.core.frame.DataFrame)


def test_silo_patch_point():
    """Test download of data from SILO patch point"""
    tmp_silo = silo.Silo(
        location=31011, 
        date_range=("20220101", "20220131"), 
        met_data=None, 
        comment="rxel",
        username="testing-glm-met@pytest.com", 
        api="patch_point"
    )

    tmp_silo.get_variables(request_settings=None)

    assert isinstance(tmp_silo.met_data.metadata, dict) and isinstance(tmp_silo.met_data.data, pd.core.frame.DataFrame)


def test_silo_drill_down_write_met():
    """Test writing data downloaded from SILO API to file"""
    tmp_silo = silo.Silo(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None,
        comment="rxel",
        username="testing-glm-met@pytest.com", 
        api="data_drill"
    )

    tmp_silo.get_variables(request_settings=None)
    
    with tempfile.TemporaryDirectory() as f:
        tmp_silo.write_met(path=f)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_raw.zip"), "r") as ff:
            f_namelist = ff.namelist()
    
    assert ("met_raw.zip" in f_files) and ("met_raw.csv" in f_namelist)


def test_silo_patch_point_write_met():
    """Test writing data downloaded from SILO API to file"""
    tmp_silo = silo.Silo(
        location=31011, 
        date_range=("20220101", "20220131"), 
        met_data=None, 
        comment="rxel",
        username="testing-glm-met@pytest.com", 
        api="patch_point"
    )

    tmp_silo.get_variables(request_settings=None)
    
    with tempfile.TemporaryDirectory() as f:
        tmp_silo.write_met(path=f)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_raw.zip"), "r") as ff:
            f_namelist = ff.namelist()
    
    assert ("met_raw.zip" in f_files) and ("met_raw.csv" in f_namelist)


import json
import os
import tempfile
import zipfile
from typing import Union

import pandas as pd
import requests

from glm_met import glm_met

from . import settings


class MetData(glm_met.MetData):
    """
    Class for meteorological data and its associated metadata.

    Used  to store meteoroligcal data downloaded from the
    NASA POWER API.

    Attributes
    ----------
    metadata : (dict)
        Metadata associated with the meteorological data.
    data : (pd.DataFrame)
        Meteorological data in a Pandas DataFrame.
    """

    def __init__(self, metadata: dict, data: pd.DataFrame):
        self.metadata = metadata
        self.data = data


class Power(glm_met.GlmMet):
    """
    Class for retrieving and processing meteorological data from
    the NASA POWER API.
    """

    # class attribute storing base URL for openmeteo historical API
    nasa_power_api_url = settings.nasa_power_api_url

    def __init__(
        self,
        location: tuple[float, float],
        date_range: tuple[str, str],
        met_data: Union[None, MetData],
        glm_met_data: Union[None, pd.DataFrame],
        parameters: Union[None, list[str]],
        timezone: str = "LST",
        community: str = "SB",
    ):
        """Initialises a `Power` object retrieving and storing
        meteorological data from NASA POWER's API.

        Parameters
        ----------
        location : tuple[float, float]
            Latitude and longitude of the location for the data retrieval.
        date_range : tuple[str, str]
            Start and end dates for the data retrieval
            (in ISO 8601 format, e.g., "YYYY-MM-DD").
        met_data : Union[None, MetData]
            Attribute to store meteorological data downloaded from open-meteo.
        glm_met_data : Union[None, pd.DataFrame]
            Attribute to store meteorological data downloaded from open-meteo
            in GLM format.
        parameters : Union[None, list[str]]
            List of meteorological variables to retrieve. The default is
            to use the list specified in `settings.hourly_glm_default`.
        timezone : str
            LST - local solar time or UTC.
        community : str
            `SB` for sustainable buildings and `AG` for agroclimatology.
        """
        self.location = location
        self.date_range = date_range
        self.met_data = met_data
        self.glm_met_data = glm_met_data
        self.timezone = timezone
        self.parameters = parameters
        self.community = community
        self.r_format = "JSON"

    def get_variables(self, request_settings: Union[None, dict]) -> MetData:
        """
        Get variables from the NASA POWER hourly API and store them
        in the `met_data` attribute.

        If you need to pass in extra settings to the API, pass these into
        the `get_variables()` method as a dict object. These settings will
        be appended to the API request as query parameters.

        Parameters
        ----------
        request_settings : Union[None, dict]
            Dictionary object of extra settings to pass in as query parameters.
        """

        # hourly
        if self.parameters is None:
            self.parameters = settings.hourly_glm_default

        payload = {
            "longitude": self.location[0],
            "latitude": self.location[1],
            "start": self.date_range[0].replace("-", ""),
            "end": self.date_range[1].replace("-", ""),
            "parameters": (",").join(self.parameters),
            "timezone": self.timezone,
            "community": self.community,
            "format": self.r_format,
        }

        # if user has supplied extra query parameters, append them here.
        if request_settings:
            payload = {**payload, **request_settings}

        r = requests.get(Power.nasa_power_api_url, params=payload)
        r_json = r.json()
        df = pd.DataFrame(r_json["properties"]["parameter"])
        metadata = {
            key: r_json[key]
            for key in r_json.keys() & {"header", "parameters"}
        }

        self.met_data = MetData(metadata=metadata, data=df)

    def write_met(self, path: str) -> None:
        """
        Save meteorological data and its metadata to a zip file.

        Parameters
        ----------
        path : str
            Path to the directory where the zip file should be saved.
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            self.met_data.data.to_csv(
                os.path.join(tmpdir, "met_raw.csv"), index=False
            )

            with open(
                os.path.join(tmpdir, "met_raw_metadata.json"), "w"
            ) as dst:
                json.dump(self.met_data.metadata, dst)

            data_fpath = os.path.join(tmpdir, "met_raw.csv")
            metadata_fpath = os.path.join(tmpdir, "met_raw_metadata.json")

            files = [data_fpath, metadata_fpath]

            with zipfile.ZipFile(
                os.path.join(path, "met_raw.zip"), mode="w"
            ) as z_dst:
                for file in files:
                    z_dst.write(file, arcname=file.split("/")[-1])

    def convert_to_glm(self) -> None:
        """
        Convert hourly meteorological data to GLM format
        and store it in the `glm_met_data` attribute.

        This function only works with hourly data, this method expects the
        `met_data.data` attribute of the `Power` object to have the
        following variables: `"ALLSKY_SFC_SW_DWN"`, `"CLOUD_AMT"`, `"T2M"`, `"RH2M"`,
        `"WS2M"`, `"PRECTOTCORR"`.

        If the data in `met_data.data` has different
        variables, you will need to write your own function to convert the data
        to GLM format.

        `CLOUD_AMT` from NASA POWER is in percentage units. GLM requires it as
        a proportion (0.0 to 1.0).

        `precipitation` is in mm. GLM requires it in m / day even if
        using hourly data, so it is converted.
        """

        df_glm = pd.DataFrame(
            # glm requires time to have "hh:mm:ss" - append :00 to date-time string
            data={
                "time": pd.to_datetime(
                    self.met_data.data.index, format="%Y%m%d%H"
                ).astype(str),
                "ShortWave": self.met_data.data["ALLSKY_SFC_SW_DWN"],
                "Cloud": self.met_data.data["CLOUD_AMT"] / 100,
                "AirTemp": self.met_data.data["T2M"],
                "RelHum": self.met_data.data["RH2M"],
                "WindSpeed": self.met_data.data["WS2M"],
                "Rain": (self.met_data.data["PRECTOTCORR"] / 1000) * 24,
            }
        ).reset_index(drop=True)

        self.glm_met_data = df_glm

    def write_glm_met(self, path: str, zip_f: bool, fname: str) -> None:
        """
        Save meteorological data in GLM format and its metadata to file.

        Only works after a call to `convert_to_glm()` and with hourly
        data.

        Parameters
        ----------
        path : str
            Path to the directory where the zip file should be saved.
        zip_f : bool
            Whether to save output to a zip file.
        fname : str
            If zip_f is `True`, filename for GLM meteorological data csv.
        """

        glm_metadata = self.met_data.metadata

        if zip_f:
            with tempfile.TemporaryDirectory() as tmpdir:
                self.glm_met_data.to_csv(
                    os.path.join(tmpdir, "met.csv"), index=False
                )

                with open(
                    os.path.join(tmpdir, "met_glm_metadata.json"), "w"
                ) as dst:
                    json.dump(glm_metadata, dst)

                data_fpath = os.path.join(tmpdir, "met.csv")
                metadata_fpath = os.path.join(tmpdir, "met_glm_metadata.json")

                files = [data_fpath, metadata_fpath]

                with zipfile.ZipFile(
                    os.path.join(path, "met_glm.zip"), mode="w"
                ) as z_dst:
                    for file in files:
                        z_dst.write(file, arcname=file.split("/")[-1])
        else:
            self.glm_met_data.to_csv(os.path.join(path, fname), index=False)



import pytest
import tempfile
import os
import pandas as pd
import zipfile

import glm_met.nasa_power.nasa_power as nasa_power

def test_np_init():
    """Test initialise Power object"""
    tmp_power = nasa_power.Power(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None,
        glm_met_data=None,
        parameters=None,
    )

    assert isinstance(tmp_power, nasa_power.Power)


def test_nasa_power_download():
    """Test download of data from NASA POWER"""
    tmp_power = nasa_power.Power(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None,
        glm_met_data=None,
        parameters=None,
    )

    tmp_power.get_variables(request_settings=None)

    assert isinstance(tmp_power.met_data.metadata, dict) and isinstance(tmp_power.met_data.data, pd.core.frame.DataFrame)


def test_nasa_power_write_met():
    """Test writing data downloaded from NASA POWER to file"""
    tmp_power = nasa_power.Power(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None,
        glm_met_data=None,
        parameters=None, 
    )

    tmp_power.get_variables(request_settings=None)
    
    with tempfile.TemporaryDirectory() as f:
        tmp_power.write_met(path=f)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_raw.zip"), "r") as ff:
            f_namelist = ff.namelist()
    
    assert ("met_raw.zip" in f_files) and ("met_raw.csv" in f_namelist)


def test_nasa_power_convert_to_glm():
    """Test converting downloaded met data to GLM format"""
    tmp_power = nasa_power.Power(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None,
        glm_met_data=None,
        parameters=None, 
    )

    tmp_power.get_variables(request_settings=None)
    tmp_power.convert_to_glm()
    src_glm_cols = tmp_power.glm_met_data.columns
    target_glm_cols = [
        "time",
        "ShortWave",
        "Cloud",
        "AirTemp",
        "RelHum",
        "WindSpeed",
        "Rain"
    ]
    matching_cols = [c for c in src_glm_cols if c  in target_glm_cols]
    
    assert (len(matching_cols) == 7) and (tmp_power.met_data.data.shape[0] == tmp_power.glm_met_data.shape[0])

def test_nasa_power_write_glm_zip():
    """Test writing GLM formatted data file - zip"""
    tmp_power = nasa_power.Power(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None,
        glm_met_data=None,
        parameters=None, 
    )

    tmp_power.get_variables(request_settings=None)
    tmp_power.convert_to_glm()
    
    with tempfile.TemporaryDirectory() as f:
        tmp_power.write_glm_met(f, True, None)
        # check zipfile in tmp dir
        f_files = os.listdir(f)
        with zipfile.ZipFile(os.path.join(f, "met_glm.zip"), "r") as ff:
            f_namelist = ff.namelist()
    
    assert ("met_glm.zip" in f_files) and ("met.csv" in f_namelist)

def test_nasa_power_write_glm():
    """Test writing GLM formatted data file"""
    tmp_power = nasa_power.Power(
        location=(116.6, -32.17), 
        date_range=("20220101", "20220131"), 
        met_data=None,
        glm_met_data=None,
        parameters=None, 
    )

    tmp_power.get_variables(request_settings=None)
    tmp_power.convert_to_glm()
    
    with tempfile.TemporaryDirectory() as f:
        tmp_power.write_glm_met(f, False, "test.csv")
        f_files = os.listdir(f)
    
    assert ("test.csv" in f_files)